import pyscf
import pyscf.sidereus
import numpy as np
import pytest
import os
import glob

class TestCDFT:
    def setup_method(self, method):
        self.atom = "H 0 0 0; F 0 0 0.917"

    def test_cdft_xtb_fmo(self):
        mol = pyscf.M(atom=self.atom)
        mf = mol.sXTB(method="GFN2-xTB")
        mf.kernel()
        
        # FMO Method
        print("\nTesting XTB FMO...")
        res = mf.sCDFT(method="fmo").kernel()
        
        # Check structure
        assert "global" in res
        assert "atomic" in res
        
        # Check Global
        glob = res["global"]
        assert "VIP" in glob
        assert "Hardness" in glob
        assert glob["VIP"] > 0
        
        # Check Atomic
        atom = res["atomic"]
        assert "f+" in atom
        assert len(atom["f+"]) == mol.natm
        
        # Check new descriptors
        assert "s+" in atom
        assert "omega+" in atom
        assert "n-" in atom

    def test_cdft_xtb_fd(self):
        mol = pyscf.M(atom=self.atom)
        mf = mol.sXTB(method="GFN2-xTB")
        mf.kernel()
        
        # FD Method
        print("\nTesting XTB FD...")
        res = mf.sCDFT(method="fd").kernel()
        
        glob = res["global"]
        atom = res["atomic"]
        
        print("FD Global:", glob)
        print("FD Atomic f+:", atom["f+"])
        
        assert glob["VIP"] > 0
        assert np.abs(np.sum(atom["f+"]) - 1.0) < 0.2
        
        assert "s-" in atom

    def test_cdft_rhf_fmo(self):
        mol = pyscf.M(atom=self.atom, basis='sto-3g')
        mf = mol.RHF()
        mf.kernel()
        
        print("\nTesting RHF FMO...")
        res = mf.sCDFT(method="fmo").kernel()
        
        glob = res["global"]
        atom = res["atomic"]
        
        assert glob["VIP"] > 0
        assert np.sum(atom["f-"]) > 0.9
        assert "omega+" in atom

    def test_cdft_rhf_fd(self):
        mol = pyscf.M(atom=self.atom, basis='sto-3g')
        mf = mol.RHF()
        mf.kernel()
        
        print("\nTesting RHF FD...")
        res = mf.sCDFT(method="fd").kernel()
        
        glob = res["global"]
        atom = res["atomic"]
        
        assert glob["VIP"] > 0
        assert np.sum(atom["f+"]) > 0.8


class TestSpatial:
    def setup_method(self, method):
        self.atom = "H 0 0 0; F 0 0 0.917"
        # Cleanup cube files before test
        for f in glob.glob("*.cube"):
            os.remove(f)

    def teardown_method(self, method):
        # Cleanup cube files after test
        for f in glob.glob("*.cube"):
            os.remove(f)

    def test_spatial_rhf_fmo(self):
        mol = pyscf.M(atom=self.atom, basis='sto-3g')
        mf = mol.RHF()
        mf.kernel()
        
        print("\nTesting RHF FMO Spatial...")
        # New API: .spatial().kernel()
        res = mf.sCDFT(method="fmo").spatial(nx=40, ny=40, nz=40).kernel()
        
        files = res
        assert files is not None
        # 4 Fukui + 4 Softness + 1 Electrophilicity + 1 Nucleophilicity = 10 files
        assert len(files) == 10
        for f in files:
            assert os.path.exists(f)
            
    def test_spatial_rhf_fd(self):
        mol = pyscf.M(atom=self.atom, basis='sto-3g')
        mf = mol.RHF()
        mf.kernel()
        
        print("\nTesting RHF FD Spatial...")
        res = mf.sCDFT(method="fd").spatial(nx=40, ny=40, nz=40).kernel()
        
        files = res
        assert files is not None
        assert len(files) == 10
        for f in files:
            assert os.path.exists(f)

    def test_spatial_xtb_warning(self):
        # Should warn and return None for spatial
        mol = pyscf.M(atom=self.atom)
        mf = mol.sXTB(method="GFN2-xTB")
        mf.kernel()
        
        print("\nTesting XTB Spatial Warning...")
        res = mf.sCDFT(method="fmo").spatial().kernel()
        
        assert res is None


from pyscf import lib, scf, dft
import numpy as np

try:
    from ..xtb import XTB
    HAS_XTB = True
except ImportError:
    HAS_XTB = False

def get_homo_lumo(mf):
    if not hasattr(mf, "mo_energy") or not hasattr(mf, "mo_occ"):
            raise RuntimeError("Orbital data missing.")

    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    
    is_uhf = False
    if isinstance(mo_energy, tuple) or (isinstance(mo_energy, np.ndarray) and mo_energy.ndim == 2 and mo_energy.shape[0] == 2):
            is_uhf = True
            
    deg_thr = 1e-4

    if is_uhf:
        if isinstance(mo_energy, tuple):
            ea_a, ea_b = mo_energy
            occ_a, occ_b = mo_occ
        else:
            ea_a, ea_b = mo_energy[0], mo_energy[1]
            occ_a, occ_b = mo_occ[0], mo_occ[1]
        
        def get_idx_e(e, o):
            e = np.asarray(e)
            o = np.asarray(o)
            occ_idx = np.where(o > 1e-6)[0]
            virt_idx = np.where(o <= 1e-6)[0]
            
            # Highest occupied energy
            h_idx = occ_idx[-1] if len(occ_idx) > 0 else None
            h_e = e[h_idx] if h_idx is not None else -np.inf
            
            # Lowest virtual energy
            l_idx = virt_idx[0] if len(virt_idx) > 0 else None
            l_e = e[l_idx] if l_idx is not None else np.inf
            
            # Check degeneracy
            h_idxs = []
            if h_idx is not None:
                # Find all occupied orbitals degenerate with HOMO
                # Note: usually we check occupied ones. 
                # e[occ_idx] check against h_e
                deg_mask = (np.abs(e[occ_idx] - h_e) < deg_thr)
                h_idxs = occ_idx[deg_mask].tolist()
                
            l_idxs = []
            if l_idx is not None:
                # Find all virtual orbitals degenerate with LUMO
                deg_mask = (np.abs(e[virt_idx] - l_e) < deg_thr)
                l_idxs = virt_idx[deg_mask].tolist()
                
            return h_idxs, l_idxs, h_e, l_e
        
        idxs_ha, idxs_la, ha, la = get_idx_e(ea_a, occ_a)
        idxs_hb, idxs_lb, hb, lb = get_idx_e(ea_b, occ_b)
        
        return {
            "is_uhf": True,
            "ha": (idxs_ha, ha), "la": (idxs_la, la),
            "hb": (idxs_hb, hb), "lb": (idxs_lb, lb),
            "homo_e": max(ha, hb), "lumo_e": min(la, lb)
        }
    else:
        e = np.asarray(mo_energy)
        o = np.asarray(mo_occ)
        occ_idx = np.where(o > 1e-6)[0]
        virt_idx = np.where(o <= 1e-6)[0]
        
        h_idx = occ_idx[-1] if len(occ_idx) > 0 else None
        h_e = e[h_idx] if h_idx is not None else -np.inf
        
        l_idx = virt_idx[0] if len(virt_idx) > 0 else None
        l_e = e[l_idx] if l_idx is not None else np.inf
        
        h_idxs = []
        if h_idx is not None:
            deg_mask = (np.abs(e[occ_idx] - h_e) < deg_thr)
            h_idxs = occ_idx[deg_mask].tolist()
            
        l_idxs = []
        if l_idx is not None:
            deg_mask = (np.abs(e[virt_idx] - l_e) < deg_thr)
            l_idxs = virt_idx[deg_mask].tolist()
        
        return {
            "is_uhf": False,
            "h": (h_idxs, h_e), "l": (l_idxs, l_e),
            "homo_e": h_e, "lumo_e": l_e
        }

def run_fd_calc(mf, charge_delta, cache=None):
    if cache is not None and charge_delta in cache:
        return cache[charge_delta]

    mol_neu = mf.mol
    mol_new = mol_neu.copy()
    mol_new.charge += charge_delta

    nelec = mol_new.nelectron
    if nelec % 2 == 1:
        mol_new.spin = 1
    else:
        mol_new.spin = 0
    mol_new.build(0, 0)

    if HAS_XTB and isinstance(mf, XTB):
        mf_new = mol_new.sXTB(method=mf.method)
        mf_new.verbose = 0
        
        from tblite.interface import Calculator
        calc = Calculator(
            method=mf_new.method,
            numbers=mol_new.atom_charges(),
            positions=mol_new.atom_coords(),
            charge=mol_new.charge,
            uhf=mol_new.spin
        )
        res = calc.singlepoint()
        e_tot = res.get("energy")
        charges = res.get("charges")
        density = None
        
        res_tuple = (e_tot, charges, density)
        if cache is not None: cache[charge_delta] = res_tuple
        return res_tuple
    else:
        mf_cls = mf.__class__
        if not getattr(mf, 'unrestricted', False) and mol_new.spin != 0:
            if issubclass(mf_cls, scf.hf.RHF): mf_cls = scf.uhf.UHF
            elif issubclass(mf_cls, dft.rks.RKS): mf_cls = dft.uks.UKS
        mf_new = mf_cls(mol_new)
        if hasattr(mf, 'xc'): mf_new.xc = mf.xc
        if hasattr(mf, 'grids'): mf_new.grids = mf.grids
        mf_new.verbose = 0
        mf_new.run()

        e_tot = mf_new.e_tot
        pop, chg = mf_new.mulliken_pop(verbose=0)
        charges = chg

        density = mf_new.make_rdm1()

        res_tuple = (e_tot, charges, density)
        if cache is not None: cache[charge_delta] = res_tuple
        return res_tuple

def compute_global_from_vip_vea(vip, vea, homo):
    chi = (vip + vea) / 2.0
    mu = -chi
    eta = vip - vea
    
    if abs(eta) > 1e-12:
        s_val = 1.0 / eta
        omega = (mu ** 2) / (2 * eta)
    else:
        s_val = np.inf
        omega = np.inf
        
    n_nu = homo
    
    return {
        "VIP": vip, "VEA": vea, "Mulliken Electronegativity": chi,
        "Chemical Potential": mu, "Hardness": eta, "Softness": s_val,
        "Electrophilicity Index": omega, "Nucleophilicity Index": n_nu
    }

def calculate_global_scalars(mf, method="fmo", cache=None):
    if method == "fmo":
        hl = get_homo_lumo(mf)
        homo, lumo = hl["homo_e"], hl["lumo_e"]
        return compute_global_from_vip_vea(-homo, -lumo, homo)
    else:
        e_neu, _, _ = run_fd_calc(mf, 0, cache)
        e_cat, _, _ = run_fd_calc(mf, 1, cache)
        e_an, _, _ = run_fd_calc(mf, -1, cache)
        
        vip = e_cat - e_neu
        vea = e_neu - e_an
        homo_proxy = -vip
        return compute_global_from_vip_vea(vip, vea, homo_proxy)



import pyscf
import pyscf.sidereus
import pytest
import os
import numpy as np

class TestTDDFT:
    def setup_method(self, method):
        # Ethylene
        self.atom = "C 0 0 0; C 0 0 1.33; H 0 0.92 -0.54; H 0 -0.92 -0.54; H 0 0.92 1.87; H 0 -0.92 1.87"
        if os.path.exists("tddft_spectrum.png"):
            os.remove("tddft_spectrum.png")

    def teardown_method(self, method):
        if os.path.exists("tddft_spectrum.png"):
            os.remove("tddft_spectrum.png")

    def test_tddft_rhf(self):
        mol = pyscf.M(atom=self.atom, basis="sto-3g")
        mf = mol.RHF()
        mf.kernel()
        
        # TDDFT
        nstates = 3
        tddft = mf.sTDDFT(nstates=nstates)
        e, f = tddft.kernel()
        
        assert len(e) == nstates
        assert len(f) == nstates
        assert np.all(e > 0)
        
        # Plot
        tddft.summary(spectrum="tddft_spectrum.png")
        assert os.path.exists("tddft_spectrum.png")
        
        print("\nExcitation Energies (eV):", e * 27.2114)
        print("Oscillator Strengths:", f)


from pyscf import lib, scf
from pyscf.hessian import thermo
import numpy as np

class THERMO(lib.StreamObject):
    def __init__(self, mf, temp=298.15, press=101325):
        self.mf = mf
        self.mol = mf.mol
        self.temp = temp
        self.press = press
        self.verbose = mf.verbose
        self.stdout = mf.stdout
        self.results = {}

    def kernel(self, hessian=None):
        if self.mf.e_tot is None:
            self.mf.kernel()

        if hessian is None:
            if self.verbose >= lib.logger.NOTE:
                lib.logger.note(self.mf, "Calculating Hessian...")
            hessian = self.mf.Hessian().kernel()
        
        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(self.mf, "Performing harmonic analysis...")
            if hasattr(hessian, "shape"):
                lib.logger.note(self.mf, f"Hessian shape: {hessian.shape}")
        ha_res = thermo.harmonic_analysis(self.mol, hessian)
        
        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(self.mf, "Calculating thermodynamic properties at T=%.2f K, P=%.2f Pa...", 
                            self.temp, self.press)
        results = thermo.thermo(self.mf, ha_res["freq_au"], self.temp, self.press)
        
        self.results = results
        self.results["vib_dict"] = ha_res
        
        # Add convenient shortcuts
        self.G = results.get("G_tot", (None, None))[0]
        self.H = results.get("H_tot", (None, None))[0]
        self.S = results.get("S_tot", (None, None))[0]
        
        if self.verbose >= lib.logger.NOTE:
            self.summary()
            
        return self.results

    def summary(self):
        thermo.dump_thermo(self.mol, self.results)

# Inject to SCF
scf.hf.SCF.sTHERMO = lib.class_as_method(THERMO)

# Handle sXTB if it exists
try:
    from .xtb import XTB
    XTB.sTHERMO = lib.class_as_method(THERMO)
except ImportError:
    pass


from pyscf import lib, scf
import numpy as np
from pyscf import tddft
from .utils.matplotlib import plot_spectrum

# Monkeypatch lib.einsum to avoid numpy 2.x incompatibility in PySCF


def _safe_einsum(*args, **kwargs):
    if '_contract' in kwargs:
        del kwargs['_contract']
    return np.einsum(*args, **kwargs)


lib.einsum = _safe_einsum


class TDDFT(lib.StreamObject):
    def __init__(self, mf, nstates=3):
        self.mf = mf
        self.nstates = nstates
        self.verbose = mf.verbose
        self.stdout = mf.stdout
        self.tddft = None
        self.energies = None
        self.oscillator_strengths = None

    def kernel(self):
        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(
                self.mf, f"Running TDDFT for {self.nstates} states...")

        # Check for gpu4pyscf
        if self.mf.__module__.startswith("gpu4pyscf"):
            self.tddft = self.mf.TDA()
        else:
            self.tddft = self.mf.TDA()

        self.tddft.nstates = self.nstates
        self.energies = self.tddft.kernel()[0]

        # Calculate Oscillator Strengths
        e = self.energies
        mol = self.mf.mol

        # Calculate transition moments
        # transition_dipole() returns (nstates, 3)
        trans_dip = self.tddft.transition_dipole()
        f = []
        for i, ei in enumerate(e):
            # ei is in Hartree
            # |<0|r|I>| in Bohr
            # f = 2/3 * E * mu^2
            mu2 = np.linalg.norm(trans_dip[i])**2
            fi = (2.0 / 3.0) * ei * mu2
            f.append(fi)

        self.oscillator_strengths = np.array(f)

        if self.verbose >= lib.logger.NOTE:
            self.summary()

        return self.energies, self.oscillator_strengths

    def summary(self, spectrum=None, w=0.2):
        log = lib.logger.new_logger(self, 2)

        energies_ev = self.energies * 27.211386
        energies_nm = 1239.84193 / energies_ev

        log.log(
            "----------------------------------------------------\n"
            " State      Energy (eV)    Wavelength (nm)    f     \n"
            "----------------------------------------------------"
        )

        for i, (e_ev, nm, f) in enumerate(zip(energies_ev, energies_nm, self.oscillator_strengths)):
            log.log("{:5d}      {:10.4f}      {:10.4f}       {:8.4f}".format(
                i+1, e_ev, nm, f))
        log.log("----------------------------------------------------\n")

        if spectrum:
            # Plot using energies in eV
            x_min = max(0, min(energies_ev) - 1.0)
            x_max = max(energies_ev) + 1.0

            # Use width w (eV)
            fig, _, _ = plot_spectrum(energies_ev, self.oscillator_strengths, w, 0, 1000,
                                      x_min, x_max,
                                      "Energy (eV)",
                                      "Oscillator Strength",
                                      "Oscillator Strength",
                                      "Absorption Spectrum",
                                      "Sticks")
            fig.savefig(spectrum)
            fig.clf()


# Inject
scf.hf.SCF.sTDDFT = lib.class_as_method(TDDFT)
try:
    from .xtb import XTB
    pass
except ImportError:
    pass



from pyscf import lib, scf
from .rhf import OPT as RHFOPT
import numpy as np
from copy import deepcopy
from ..utils import rdkit as rdkit_utils

class SCAN(lib.StreamObject):
    def __init__(self, mf, scans, **kwargs):
        self.mf = mf
        self.mol = mf.mol
        self.scans = scans
        self.verbose = mf.verbose
        self.stdout = mf.stdout
        self.kwargs = kwargs
        self.scan_results = {
            "energies": [],
            "geometries": [],
            "coords": [] 
        }

    def kernel(self):
        return self.run_scan()

    def run_scan(self):
        steps = self.scans[0].get("steps", 10)
        for s in self.scans:
            if s.get("steps", 10) != steps:
                raise ValueError("All scan coordinates must have the same number of steps.")
        
        num_points = steps + 1
        
        scan_values = []
        for s in self.scans:
            start = s.get("start")
            stop = s.get("stop")
            if start is None or stop is None:
                 raise ValueError("Scan start and stop must be defined.")
            vals = np.linspace(start, stop, num_points)
            scan_values.append(vals)
            
        # Ensure RDKit availability
        try:
            from rdkit.Chem import rdMolTransforms
        except ImportError:
            raise ImportError("RDKit is required for geometry manipulation in SCAN.")

        for i in range(num_points):
            if self.verbose >= lib.logger.NOTE:
                lib.logger.note(self.mf, f"Scan step {i+1}/{num_points}")
            
            # Modify Geometry
            rd_mol = rdkit_utils.pyscf2rdmol(self.mf.mol)
            conf = rd_mol.GetConformer()
            
            current_vals = []
            current_constraints = []
            
            for j, s in enumerate(self.scans):
                val = float(scan_values[j][i])
                current_vals.append(val)
                atoms = s["atoms"]
                type_ = s["type"].lower()
                
                # Update RDKit Geometry
                if type_ == "bond":
                    rdMolTransforms.SetBondLength(conf, atoms[0], atoms[1], val)
                elif type_ == "angle":
                    rdMolTransforms.SetAngleDeg(conf, atoms[0], atoms[1], atoms[2], val)
                elif type_ == "dihedral":
                    rdMolTransforms.SetDihedralDeg(conf, atoms[0], atoms[1], atoms[2], atoms[3], val)
                
                # Map type to pysisyphus terminology
                pysis_type_map = {
                    "bond": "bond",
                    "stretch": "bond",
                    "angle": "bend",
                    "bend": "bend",
                    "dihedral": "torsion",
                    "torsion": "torsion"
                }
                pysis_type = pysis_type_map.get(type_, type_)
                
                # Constraint for pysisyphus: [type, atom1, atom2, ...]
                constr_list = [pysis_type] + list(atoms)
                current_constraints.append(constr_list)

            # Update PySCF mol
            new_coords = rdkit_utils.get_coords(rd_mol)
            rdkit_utils.set_coords(self.mf.mol, new_coords)
            
            # Run Constrained Optimization
            opt = self.mf.sOPT(constraints=current_constraints, **self.kwargs)
            mol_opt = opt.kernel()
            
            # Store results
            # Force recalculation of energy for the optimized geometry
            e = self.mf.kernel()
                
            self.scan_results["energies"].append(e)
            self.scan_results["geometries"].append(mol_opt.atom_coords())
            self.scan_results["coords"].append(current_vals)
            
        return self.scan_results

# Register to SCF and XTB
scf.hf.SCF.sSCAN = lib.class_as_method(SCAN)
try:
    from ..xtb import XTB
    XTB.sSCAN = lib.class_as_method(SCAN)
except ImportError:
    pass



import pyscf
import pyscf.sidereus


class TestXTB:
    def setup_method(self, method):
        self.atom = "H 0 0 0; F 0 0 0.917"

    def test_xtb(self):
        mol = pyscf.M(atom=self.atom)
        mf = mol.sXTB(method="GFN2-xTB")
        mf.kernel()
        mf.Gradients().kernel()

    def test_opt(self):
        mol = pyscf.M(atom=self.atom)
        mol.sXTB(method="GFN2-xTB").sOPT().kernel()